from typing import Union, Callable, Awaitable
from jsonpatch import JsonPatch
import websockets
import asyncio
import inspect
import logging
import json

from ..models import Kettle
_LOGGER = logging.getLogger(__name__)


class HWSocket:
    def __init__(self, token: str):
        self.token = token
        self.devices = {}
        self.message_id = 1
        self.message_id_futures = {}
        self.device_update_callbacks = {}
        self.ws = None

    def _get_device(self, device_id: str) -> Kettle:
        device_dict = self.devices[device_id]
        d_type = device_dict.get("type")

        if d_type == "kettle":
            return Kettle(device_dict)

    async def _on_response_msg(self, msg: dict):
        message_id = msg.get("message_id")
        if message_id not in self.message_id_futures.keys():
            _LOGGER.error(f"message_id {message_id} not found", msg)
            return

        self.message_id_futures[message_id](msg)

    async def _on_device_msg(self, msg: dict):
        device_id = msg.get("device")
        self.devices[device_id] = msg
        await self._notify_device_handler(device_id)

    async def _on_device_update_msg(self, msg: dict):
        device_id = msg.get("device")

        patch = JsonPatch(msg.get("patch", []))
        patch.apply(self.devices[device_id], in_place=True)

        await self._notify_device_handler(device_id)

    async def _on_device_name_msg(self, msg: dict):
        device_id = msg.get("device")
        name = msg.get("name")

        device = self._get_device(device_id)
        device.set_name(name)
        self.devices[device_id] = device.to_json()

        await self._notify_device_handler(device_id)

    async def _notify_device_handler(self, device_id: str):
        device = self._get_device(device_id)

        if device_id in self.device_update_callbacks:
            fn = self.device_update_callbacks[device_id]

            if inspect.iscoroutinefunction(fn):
                await self.device_update_callbacks[device_id](device)
            else:
                fn(device)

    async def subscribe_device(self, device_id: str, callback: Callable[[Kettle], Awaitable[None]]):
        _LOGGER.info(f"Subscribing to device {device_id}")
        self.device_update_callbacks[device_id] = callback

        future = asyncio.get_event_loop().create_future()
        self.device_update_callbacks[device_id] = future.set_result

        await self.send("subscribe_device", {
            "device": device_id
        })

        kettle = await future
        self.device_update_callbacks[device_id] = callback
        await callback(kettle)

        return kettle

    async def _on_message(self, message: str):
        _LOGGER.info(f"_on_message(message={repr(message)})")
        msg = json.loads(message)
        msg_type_handlers = {
            "response": self._on_response_msg,
            "kettle": self._on_device_msg,
            "device_name": self._on_device_name_msg,
            "json_patch": self._on_device_update_msg,
        }

        msg_type = msg.get("type")
        if msg_type not in msg_type_handlers.keys():
            _LOGGER.error(f"Message type {msg_type} not found", msg)
            return

        await msg_type_handlers[msg_type](msg)

    async def connect(self, callback: Callable[[], Awaitable[None]]):
        await asyncio.gather(
            self._connect(callback),
            self._receive_messages()
        )

    async def reconnect(self):
        _LOGGER.info("Reconnecting")
        future = asyncio.get_event_loop().create_future()
        callback = lambda: (future.set_result(True))
        await self._connect(callback)
        await asyncio.wait_for(future, timeout=60)

        for device_id, callback in self.device_update_callbacks.copy().items():
            await self.subscribe_device(device_id, callback)

    async def _connect(self, callback: Callable[[], Union[Awaitable[None], None]]):
        """Initiate connection"""
        self.devices = {}
        self.message_id = 1
        self.message_id_futures = {}
        self.device_update_callbacks = {}
        self.ws = await websockets.connect("wss://app-ws.homewizard.com/ws")

        await self.send("hello", {
            "type": "hello",
            "os": "android",
            "source": "kitchen",
            "version": "1.4.4",
            "token": self.token,
            "compatibility": 2
        })

        if inspect.iscoroutinefunction(callback):
            await callback()
        else:
            callback()

    async def _receive_messages(self):
        """
        Loop to keep getting websocket messages coming in
        """
        while True:
            await asyncio.sleep(0.5)
            if not self.ws:
                continue
            try:
                message = await self.ws.recv()
                asyncio.get_event_loop().create_task(self._on_message(message))
            except websockets.exceptions.ConnectionClosed:
                _LOGGER.error("_receive_messages(): Connection lost")
                await self.reconnect()
                break

    async def send(self, msg_type: str, payload: dict):
        message_id = self.message_id
        self.message_id += 1

        future = asyncio.get_event_loop().create_future()
        self.message_id_futures[message_id] = future.set_result

        try:
            await self.ws.send(json.dumps({
                "device": None,
                "model": None,
                "online": None,
                "state": None,
                **payload,
                "type": msg_type,
                "message_id": message_id
            }))
        except websockets.ConnectionClosedError:
            _LOGGER.error("send(): Connection lost")
            await self.reconnect()
            return await self.send(msg_type, payload)

        return await future

    async def update(self, device: Kettle):
        device_id = device.get_id()
        updated = device.to_json()
        patch = JsonPatch.from_diff(self.devices[device_id], updated)

        return await self.send("json_patch", {
            "device": device_id,
            "patch": patch.patch
        })



from ..enums import KettleStatus
import collections
import copy


class Kettle:
    def __init__(self, device: dict) -> None:
        self.device = copy.deepcopy(device)

    def __deep_update(self, orig_dict: dict, new_dict: dict) -> dict:
        for key, val in new_dict.items():
            if isinstance(val, collections.abc.Mapping):
                orig_dict[key] = self.__deep_update(orig_dict.get(key, {}), val)
            elif isinstance(val, list):
                orig_dict[key] = (orig_dict.get(key, []) + val)
            else:
                orig_dict[key] = new_dict[key]
        return orig_dict

    def _deep_update(self, new_dict: dict):
        return self.__deep_update(self.device, new_dict)

    def to_json(self) -> dict:
        return copy.deepcopy(self.device)

    def start(self):
        self._deep_update({"state": {
            "action": "start"
        }})

    def stop(self):
        self._deep_update({"state": {
            "action": "stop"
        }})

    def set_target_temperature(self, target_temperature: int):
        # should be between 40 and 100 in steps of 5
        self._deep_update({"state": {
            "target_temperature": target_temperature
        }})

    def get_target_temperature(self) -> int:
        return self.device.get("state", {}).get("target_temperature")

    def get_current_temperature(self) -> int:
        return self.device.get("state", {}).get("current_temperature")

    def get_status(self) -> KettleStatus:
        return KettleStatus(self.device.get("state", {}).get("status"))

    def is_online(self) -> bool:
        return self.device.get("online", False)

    def get_version(self) -> str:
        return self.device.get("version")

    def get_model(self) -> str:
        return self.device.get("model")

    def get_id(self) -> str:
        return self.device.get("device")

    def set_name(self, name: str):
        self._deep_update({"name": name})

    def get_name(self) -> str:
        return self.device.get("name")

    def get_boil_before_target(self) -> bool:
        return self.device.get("state", {}).get("boil_before_target", False)

    def set_boil_before_target(self, bbt: bool):
        self.device.setdefault("state", {})["boil_before_target"] = bbt

    def get_keep_warm_enabled(self) -> bool:
        return self.device.get("state", {}).get("keep_warm_enabled", False)

    def set_keep_warm_enabled(self, kw: bool):
        self.device.setdefault("state", {})["keep_warm_enabled"] = kw

    def get_keep_warm_set_time(self) -> int:
        return self.device.get("state", {}).get("keep_warm_set_time", False)

    def set_keep_warm_set_time(self, kw_time: int):
        self.device.setdefault("state", {})["keep_warm_set_time"] = kw_time



import urllib.parse
import requests
import hashlib
import base64


class HWRestClient:
    def __init__(self, username: str, password: str):
        self.endpoint = "https://api.homewizardeasyonline.com/v1/"
        self.username = username
        self.password = password

    def auth_get(self, resource: str) -> dict:
        url = urllib.parse.urljoin(self.endpoint, resource)
        encoded_pass = self.password.encode('utf-8')
        sha1_pass = hashlib.sha1(encoded_pass).hexdigest()

        encoded_token = f"{self.username}:{sha1_pass}".encode("utf-8")
        base64_token = base64.b64encode(encoded_token).decode("utf-8")

        return requests.get(url, headers={
            "Authorization": f"Basic {str(base64_token)}"
        }).json()

    def login(self):
        self.token = self.auth_get("auth/login").get("token")

    def get_account(self):
        return self.auth_get("account")

    def get_devices(self):
        return self.auth_get("auth/devices")



from enum import Enum


class KettleStatus(Enum):
    ON = "on"
    HEATING_TO_TARGET = "heating_to_target"
    FINISHED_HEATING_TO_TARGET = "finished_heating_to_target"
    KEEPING_WARM = "keeping_warm"
